@@ -1,5 +1,5 @@
 ;;  -*-  indent-tabs-mode:nil; coding: utf-8 -*-
-;;  Copyright (C) 2014,2015
+;;  Copyright (C) 2014,2015,2016,2017
 ;;      "Mu Lei" known as "NalaGinrut" <NalaGinrut@gmail.com>
 ;;  Artanis is free software: you can redistribute it and/or modify
 ;;  it under the terms of the GNU General Public License and GNU
@@ -33,9 +33,11 @@
   #:use-module (artanis env)
   #:use-module (artanis mime)
   #:use-module (artanis upload)
+  #:use-module (artanis websocket)
   #:use-module (artanis third-party json)
   #:use-module (artanis third-party csv)
-  #:use-module (ice-9 regex) ; FIXME: should use irregex!
+  #:use-module (artanis server scheduler)
+  #:use-module (artanis irregex)
   #:use-module (ice-9 match)
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-9)
@@ -64,7 +66,8 @@
             :mime
             :auth
             :session
-            :from-post))
+            :from-post
+            :websocket))
 
 (define (define-handler method rule opts-and-handler)
   (let ((keys (rule->keys rule))
@@ -72,31 +75,30 @@
         (opts (oah->opts opts-and-handler))
         (handler (oah->handler opts-and-handler)))
     (hash-set! *handlers-table*
-               (string-append method " " path-regexp)
+               (cons method path-regexp)
                (make-handler-rc handler keys (new-oht opts #:rule rule #:keys keys)))))
 
-(define (get rule . opts-and-handler) (define-handler "GET" rule opts-and-handler))
-(define (post rule . opts-and-handler) (define-handler "POST" rule opts-and-handler))
-(define (put rule . opts-and-handler) (define-handler "PUT" rule opts-and-handler))
-(define (patch rule . opts-and-handler) (define-handler "PATCH" rule opts-and-handler))
+(define (get rule . opts-and-handler) (define-handler 'GET rule opts-and-handler))
+(define (post rule . opts-and-handler) (define-handler 'POST rule opts-and-handler))
+(define (put rule . opts-and-handler) (define-handler 'PUT rule opts-and-handler))
+(define (patch rule . opts-and-handler) (define-handler 'PATCH rule opts-and-handler))
 ;; NOTE: delete method is rarely used, and this name will override the list deletion in Scheme.
 ;;       So my vote is changing its name, feel free to let me know if it's improper.
-(define (page-delete rule . opts-and-handler) (define-handler "DELETE" rule opts-and-handler))
+(define (page-delete rule . opts-and-handler) (define-handler 'DELETE rule opts-and-handler))
 
 ;; NOTE: we banned "\" in the path to avoid SQL-injection!!!
-(define *rule-regexp* (make-regexp ":[^\\/]+"))    
-(define *path-keys-regexp* (make-regexp ":([^\\/\\.]+)"))
+(define *rule-regexp* (string->irregex ":[^\\/]+"))
+(define *path-keys-regexp* (string->irregex ":([^\\/\\.]+)"))
 
 (define (compile-rule rule)
-  (string-append "^" 
-                 (regexp-substitute/global 
-                  #f *rule-regexp* rule 'pre "([^\\/\\?]+)" 'post)
-                 "[^ $]?$"))
+  (string-append "^"
+                 (irregex-replace/all *rule-regexp* rule "([^\\/\\?]+)")
+                 "$"))
 
 ;; parse rule-string and generate the regexp to parse keys from path-string
 (define (rule->keys rule)
-  (map (lambda (m) (match:substring m 1))
-       (list-matches *path-keys-regexp* rule)))
+  (map (lambda (m) (irregex-match-substring m 1))
+       (artanis-list-matches *path-keys-regexp* rule)))
 
 (define-syntax-rule (get-oht rc)
   (handler-rc-oht (get-handler-rc (rc-rhk rc))))
@@ -111,9 +113,7 @@
     (if h (h rc args ...) values)))
 
 (define-syntax-rule (auth-enabled? rc)
-  (begin
-    (format #t "adfa~%")
-    (hash-ref (get-oht rc) #:auth)))
+  (hash-ref (get-oht rc) #:auth))
 ;; --------------------------------------------------------------
 ;; oht handlers
 
@@ -141,7 +141,8 @@
         ('all (DB-get-all-rows r))
         ('top (DB-get-top-row r))
         ((? positive? n) (DB-get-n-rows r n))
-        (else (throw 'artanis-err 500 "raw-sql: Invalid mode!" mode))))))
+        (else (throw 'artanis-err 500 raw-sql-maker
+                     "Invalid mode!" mode))))))
 
 ;; for #:cache
 (define (cache-maker pattern rule keys)
@@ -164,7 +165,8 @@
        (try-to-cache-static-file rc file 'private maxage)))
     ((or (? boolean? opts) (? list? opts))
      (lambda (rc body) (try-to-cache-body rc body opts)))
-    (else (throw 'artanis-err "cache-maker: invalid pattern!" pattern))))
+    (else (throw 'artanis-err 500 cache-maker
+                 "Invalid pattern!" pattern))))
 
 ;; for #:cookies
 ;; NOTE: Support cookies customization, which let users define special cookies.
@@ -178,7 +180,7 @@
   (define ckl '())
   (define (cget ck)
     (or (assoc-ref ckl ck)
-        (throw 'artansi-err 500 "Undefined cookie name" ck)))
+        (throw 'artansi-err 500 cget "Undefined cookie name" ck)))
   (define (cset! ck k v)
     (and=> (cget ck) (cut %cookie-set! <> k v)))
   (define (cref ck k)
@@ -189,12 +191,12 @@
     (apply %cookie-modify (cget ck) kargs))
   (define (remove rc k)
     (let ((cookies (rc-set-cookie rc)))
-      (rc-set-cookie! 
+      (rc-set-cookie!
        rc
        `(,@cookies ,(%new-cookie #:npv '((key "")) #:expires *the-remove-expires*)))))
   (define-syntax-rule (init-cookies names)
-    (for-each 
-     (lambda (ck) 
+    (for-each
+     (lambda (ck)
        (set! ckl (cons (cons ck (new-cookie)) ckl)))
      names))
   (match mode
@@ -205,7 +207,8 @@
      (set! %cookie-ref getter)
      (set! %cookie-modify modifier)
      (init-cookies names))
-    (else (throw 'artansi-err 500 "cookies-maker: Invalid mode while init!" mode)))
+    (else (throw 'artansi-err 500 cookies-maker
+                 "Invalid mode while init!" mode)))
   (lambda (rc op)
     (case op
       ((set) cset!)
@@ -213,38 +216,41 @@
       ((setattr) setattr)
       ((update) update)
       ((remove) remove)
-      (else (throw 'artanis-err 500 "cookies-maker: Invalid operation!" op)))))
+      (else (throw 'artanis-err 500 cookies-maker
+                   "Invalid operation!" op)))))
 
 ;; for #:mime
 (define (mime-maker type rule keys)
-  (define mime (mime-guess type))
-  (lambda (rc . args)
-    (define headers `((content-type . (,(mime-guess type)))))
-    (define-syntax-rule (-> func) (values (apply func args) #:pre-headers headers))
-    (case type
-      ((json) (-> scm->json-string))
-      ((xml) (-> sxml->xml-string))
-      ((csv) (-> sxml->csv-string))
-      ((sxml) (values (object->string (car args)) #:pre-headers headers))
-      (else (throw 'artanis-err 500 "mime-maker: Invalid type!" type)))))
+  (define headers `((content-type . (,(mime-guess type)))))
+  (define-syntax-rule (-> func the-args) (values (apply func the-args) #:pre-headers headers))
+  (define gen-mime
+   (case type
+     ((json jsonp) (lambda (args) (-> ->json-string args)))
+     ((xml) (lambda (args) (-> sxml->xml-string args)))
+     ((csv) (lambda (args) (-> sxml->csv-string args)))
+     ((sxml) (lambda (args) (values (object->string (car args)) #:pre-headers headers)))
+     (else (throw 'artanis-err 500 mime-maker "Invalid type!" type))))
+  (lambda (rc . args) (gen-mime args)))
 
 ;; for #:session
 (define (session-maker mode rule keys)
-  (define* (%spawn rc #:key (keep? #f) (idname "sid") (proc session-spawn))
+  (define* (%spawn rc #:key (idname "sid") (proc session-spawn))
     (call-with-values
         (lambda () (proc rc))
       (lambda (sid session)
-        (when keep?
-          (let ((cookie (new-cookie #:npv `((,idname ,sid)))))
-            (and cookie (rc-set-cookie! rc (list cookie))))))))
-  (define* (spawn-handler rc #:key (keep? #f))
+        (let ((cookie (new-cookie #:npv `((,idname . ,sid)))))
+          (and cookie (rc-set-cookie! rc (list cookie)))
+          sid))))
+  (define spawn-handler
     (match mode
-      ((or #t 'spawn) (%spawn rc #:keep? keep?))
-      (`(spawn ,sid) (%spawn rc #:idname sid #:keep? keep?))
-      (`(spawn ,sid ,proc) (%spawn rc #:idname sid #:proc proc #:keep? keep?))
-      (else (throw 'artanis-err 500 "session-maker: Invalid config mode" mode))))
+      ((or #t 'spawn) %spawn)
+      (`(spawn ,sid) (lambda (rc) (%spawn rc #:idname sid)))
+      (`(spawn ,sid ,proc) (lambda (rc) (%spawn rc #:idname sid #:proc proc)))
+      (else (throw 'artanis-err 500 session-maker "Invalid config mode: ~a" mode))))
   (define (check-it rc idname)
-    (get-session (cookie-ref (rc-cookie rc) idname)))
+    (let ((sid (any (lambda (c) (cookie-ref c idname)) (rc-cookie rc))))
+      (and=> (session-restore (or sid ""))
+             (lambda (s) (session-from-correct-client? s rc)))))
   (lambda (rc cmd)
     (match cmd
       ('check (check-it rc "sid"))
@@ -253,21 +259,15 @@
        (or (check-it rc "sid") (spawn-handler rc)))
       (`(check-and-spawn ,sid)
        (or (check-it rc sid) (spawn-handler rc)))
-      (`(check-and-spawn-and-keep ,sid)
-       (or (check-it rc sid) (spawn-handler rc #:keep? #t)))
       ('spawn (spawn-handler rc))
-      ('spawn-and-keep (spawn-handler rc #:keep? #t))
-      (else (throw 'artanis-err 500 "session-maker: Invalid call cmd" cmd)))))
+      (else (throw 'artanis-err 500 session-maker "Invalid call cmd: ~a" cmd)))))
 
+;; for #:from-post
 (define (from-post-maker mode rule keys)
-  (define post-data #f)
   (define* (post->qstr-table rc #:optional (safe? #f))
-    (when (not post-data)
-      (set! post-data
-            (if (rc-body rc)
-                (generate-kv-from-post-qstr (rc-body rc) #:no-evil? safe?)
-                '())))
-    post-data)
+    (if (rc-body rc)
+        (generate-kv-from-post-qstr (rc-body rc) #:no-evil? safe?)
+        '()))
   (define (default-success-ret size-list filename-list)
     (call-with-output-string
      (lambda (port)
@@ -276,18 +276,23 @@
                  size-list filename-list))))
   (define (default-no-file-ret) "<p>No uploaded files!</p>")
   (define* (store-the-bv rc #:key (uid 33) (gid 33) (path (get-conf '(upload path)))
-                         (mod #o664) (path-mode #o775) (sync #f)
+                         (mode #o664) (path-mode #o775) (sync #f) (simple-ret? #t)
+                         (need-mfd? #f)
                          ;; NOTE: One may use these two things for returning
                          ;;       customized result, like JSON or XML.
                          (success-ret default-success-ret)
                          (no-file-ret "No uploaded files!"))
-    (match (store-uploaded-files rc #:path path #:sync sync #:simple-ret? #f
-                                 #:uid uid #:gid gid #:path-mode path-mode)
+    (match (store-uploaded-files rc #:path path #:sync sync #:simple-ret? simple-ret?
+                                 #:mode mode #:uid uid #:gid gid #:path-mode path-mode
+                                 #:need-mfd? need-mfd?)
+      ('success 'success)
+      ((? is-mfds? mfds) mfds)
       (`(success ,slist ,flist) (success-ret slist flist))
+      (`(success ,mfds ,slist ,flist) (success-ret mfds slist flist))
       ('none (no-file-ret))
       (else
-       (throw 'artanis-err 500
-              "store-the-bv: Impossible status! please report bug!"))))
+       (throw 'artanis-err 500 store-the-bv
+              "Impossible status! please report bug!"))))
   (define (post-ref plst key) (and=> (assoc-ref plst key) car))
   (define (post-handler rc)
     (match mode
@@ -295,19 +300,85 @@
       ('qstr-safe (post->qstr-table rc 'safe))
       ((or 'bv 'bytevector) (rc-body rc))
       ;; upload operation, indeed
-      (`(store ,path) (store-the-bv rc #:path path))
-      (`(store ,path ,mod)
-       (store-the-bv rc #:path path #:mod mod))
-      (else (throw 'artanis-err 500 "post-handler: Invalid mode!" mode))))
+      (('store rest ...)
+       (catch 'system-error
+         (lambda ()
+           (apply store-the-bv rc rest))
+         (lambda e
+           (let ((errno (system-error-errno e)))
+             (cond
+              ((= errno ENOMEM)
+               ;; NOTE: Out of memory, call (gc) and throw 507
+               (format (artanis-current-output) "No memory! Run GC now!~%")
+               (gc)
+               (throw 'artanis-err 507 post-handler
+                      "Server is out of RAMs, please extend more RAMs!~%"))
+              ((= errno EIO)
+               ;; NOTE: The storage device was disconnected, just throw 507
+               (throw 'artanis-err 507 post-handler
+                      "Server is not available, maybe storage media was disconnected?~%"))
+              ((= errno ENOSPC)
+               ;; NOTE: no space for uploading, just throw 507
+               (throw 'artanis-err 507 post-handler
+                      "Server has insufficient storage space!~%"))
+              (else
+               ;; nothing noticed, re-throw it to next level.
+               (apply throw e)))))))
+      (else (throw 'artanis-err 500 post-handler "Invalid mode `~a'!" mode))))
+  (define (get-values-from-post pl . keys)
+    (apply
+     values
+     (map
+      (lambda (k) (post-ref pl k))
+      keys)))
   (lambda (rc . cmd)
     (match cmd
       (`(get ,key) (post-ref (post-handler rc) key))
-      ('(get) (rc-body rc))
+      ('(get) (post-handler rc))
+      (('get-vals keys ...) (apply get-values-from-post (post-handler rc) keys))
       ('(store) (post-handler rc))
-      (else (throw 'artanis-err 500 "from-post-maker: Invalid cmd!" cmd)))))
+      (else (throw 'artanis-err 500 from-post-maker "Invalid cmd `~a'!" cmd)))))
+
+;; for #:websocket
+;;
+(define (websocket-maker mode rule keys)
+  (match mode
+    (((or #t 'raw))
+     (this-rule-enabled-websocket! rule 'raw))
+    (('proto (? symbol? proto))
+     ;; TODO: call protocol initilizer, and establish websocket for it.
+     ;; NOTE: In Artanis, we accept only one protocol for each URL-remapping,
+     ;;       if you have several protocols to service, please use different
+     ;;       URL-remapping.
+     (this-rule-enabled-websocket! rule proto))
+    (('redirect (? string? ip/usk))
+     ;; NOTE: We use IP rather than hostname, since it's usually redirected to
+     ;;       a LAN address. Using hostname may cause DNS issues.
+     ;; NOTE: ip/usk means ip or unix-socket, the pattern should be this:
+     ;;       ^ip://(?:[0-9]{1,3}\\.){3}[0-9]{1,3}(:[0-9]{1,5})?$
+     ;;       ^unix://[a-zA-Z-_0-9]+\\.socket$
+     ;; NOTE: Equivalent to transparent proxy.
+     ;; TODO: call protocol initilizer, and establish websocket
+     ;;       to redirect it.
+     ;; NOTE: Just call :websocket as the handler is enough to redirect data automatically
+     (this-rule-enabled-websocket! rule 'redirect))
+    (('proxy (? symbol? proto))
+     ;; NOTE: Setup a proxy with certain protocol handler.
+     ;;       Different from the regular proxy design, the proxy in Artanis doesn't
+     ;;       need a listen port, since it's always 80/443. The client should
+     ;;       support websocket, and visit the related URL for establishing
+     ;;       a websocket channel. Then the rest is the same with regular proxy.
+     (this-rule-enabled-websocket! rule 'proxy))
+    (else (throw 'artanis-err 500 websocket-maker "Invalid type `~a'!" mode)))
+  (lambda (rc . cmd)
+    ;; NOTE: :websocket is not for read/write, it's only used for changing status.
+    (match cmd
+      ('(payload) (websocket-frame-payload (rc-body rc)))
+      ('(frame) (rc-body rc))
+      (else (throw 'artanis-err 500 websocket-maker "Invalid cmd `~a'!" cmd)))))
 
 ;; ---------------------------------------------------------------------------------
-  
+
 ;; NOTE: these short-cut-maker should never be exported!
 ;; ((handler arg rule keys) rc . args)
 ;; NOTE: If the keyword wasn't specified while defining the url-remap,
@@ -315,6 +386,8 @@
 ;; CONVENTION:
 ;; 1. When a var is arounded with ${..}, and there's ":" as its prefix,
 ;;    say, ${:xxx}, it means "to get xxx from rule key".
+;; 2. Arounded with ${..}, and there's "@" as its prefix, say, ${@xxx},
+;;    it means "to get xxx from body" (POST way).
 (define *options-meta-handler-table*
   `(;; a short way for sql-mapping from the bindings in rule
     ;; e.g #:sql-mapping "select * from table where id=${:id}"
@@ -329,7 +402,7 @@
     ;; 1. SQL as string template.
     ;; 2. (table-name username-field passwd-field crypto-proc)
     ;; 3. (table-name crypto-proc), so passwd field will be "passwd" in default.
-    ;; e.g (get "/auth" 
+    ;; e.g (get "/auth"
     ;;      #:auth "select ${passwd} from blog where usrname=${myname}"
     ;;      (lambda (rc)
     ;;       (:auth rc #:usrname (params rc "username") #:passwd (params rc "passwd"))
@@ -363,7 +436,7 @@
     ;;       #:raw-sql "select * from Persons where name='nala'"
     ;;       (lambda (rc)
     ;;        (:raw-sql rc 'all)))
-    ;; It's used for a quick query, but you can't modify/specify query on the fly. 
+    ;; It's used for a quick query, but you can't modify/specify query on the fly.
     ;; Usage:
     ;;  :raw-sql procedure accepts three modes:
     ;; 1. 'all for getting all results.
@@ -375,7 +448,7 @@
     ;; The default value is #f.
     ;; This is useful for web caching.
     ;; values:
-    ;; * #t: will store the md5 of this page content, 
+    ;; * #t: will store the md5 of this page content,
     ;;       and compare it with Etag.
     ;; * #f: won't store md5, and won't do any caching work.
     ;;       Etag would be ignored. If the current rule is for dynamic page,
@@ -426,7 +499,31 @@
     (#:session . ,session-maker)
 
     ;; Get data from POST
-    (#:from-post . ,from-post-maker)))
+    (#:from-post . ,from-post-maker)
+
+    ;; Establish websocket channel
+    ;; Each time developers set :websocket and specifed a protocol, say, `myproto',
+    ;; Artanis will check if the file app/protocols/myproto.scm exists, then load
+    ;; the protocol initialize function in `myproto' which will also add `myproto'
+    ;; into *proto-table*. Then ragnarok will take charge of it to call the correct
+    ;; handlers.
+    ;; There're 3 cases:
+    ;; 1. Regular protocol handling over websocket
+    ;;    The initializer will create an instance for the protocol.
+    ;; 2. Redirect to another address (IP or UNIX socket)
+    ;;    The initializer will create a binding pair in redirect table:
+    ;;    src port and des port.
+    ;;    In this case, there's no protocol handling in Artanis. The protocol handler
+    ;;    is in remote (could be a host machine, or a process in current OS).
+    ;; 3. Proxy mode:
+    ;;    Different from reguar proxy, the proxy in Artanis runs over websocket, and
+    ;;    there's no extra listenning port, only 80/443.
+    ;;    In this case, there're 2 situations:
+    ;;    a. transparent proxy
+    ;;       No protocol handling at all, just redirecting the data without cooking.
+    ;;    b. cooked proxy
+    ;;       There's a bound protocol instance for it.
+    (#:websocket . ,websocket-maker)))
 
 (define-macro (meta-handler-register what)
   `(define-syntax-rule (,(symbol-append ': what) rc args ...)
@@ -444,6 +541,7 @@
 (meta-handler-register auth)
 (meta-handler-register session)
 (meta-handler-register from-post)
+(meta-handler-register websocket)
 
 (define-syntax-rule (:cookies-set! rc ck k v)
   ((:cookies rc 'set) ck k v))
